/* Copyright 2019-2020 National Research Foundation (SARAO)
 *
 * This program is free software: you can redistribute it and/or modify it under
 * the terms of the GNU Lesser General Public License as published by the Free
 * Software Foundation, either version 3 of the License, or (at your option) any
 * later version.
 *
 * This program is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more
 * details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

/**
 * @file
 *
 * This file is automatically generated. Do not edit.
 */

#ifndef SPEAD2_COMMON_LOADER_RDMACM_H
#define SPEAD2_COMMON_LOADER_RDMACM_H

#ifndef _GNU_SOURCE
# define _GNU_SOURCE
#endif
#include <spead2/common_features.h>

#if SPEAD2_USE_IBV

#include <rdma/rdma_cma.h>

#undef rdma_bind_addr
#undef rdma_create_event_channel
#undef rdma_create_id
#undef rdma_destroy_event_channel
#undef rdma_destroy_id

namespace spead2
{

extern int (*rdma_bind_addr)(struct rdma_cm_id *id, struct sockaddr *addr);
extern struct rdma_event_channel *(*rdma_create_event_channel)(void);
extern int (*rdma_create_id)(struct rdma_event_channel *channel, struct rdma_cm_id **id, void *context, enum rdma_port_space ps);
extern void (*rdma_destroy_event_channel)(struct rdma_event_channel *channel);
extern int (*rdma_destroy_id)(struct rdma_cm_id *id);

} // namespace spead2

#endif // SPEAD2_USE_IBV
#endif // SPEAD2_COMMON_LOADER_RDMACM_H
